# -*- coding: utf-8 -*-
from abc import ABCMeta
from collections import Counter
from itertools import combinations, combinations_with_replacement, product
from typing import Any, Dict, Iterable, List, Tuple, Union

import networkx as nx
import numpy as np

from .utils.functions import (
    chunks,
    distribute_function,
    flatten_lists,
    generate_random_strings,
)
from .utils.validation import check_wl_input


class WeisfeilerLehmanKernel:
    """Weisfeiler-Lehmann kernel"""

    def __init__(
        self,
        n_jobs: int = 4,
        precomputed: bool = False,
        n_iter: int = 3,
        node_label: str = "residue",
        normalize: bool = False,
        biased: bool = False,
        verbose: bool = False,
    ):
        self.n_iter = n_iter
        self.node_label = node_label
        self.normalize = normalize
        if n_jobs is not None:
            self.n_jobs = int(n_jobs)
        else:
            self.n_jobs = None
        self.biased = biased
        self.verbose = verbose
        self.precomputed = precomputed

    def compute_wl_hashes(self, G: nx.Graph) -> Dict:
        """Computes Weisfeiler-Lehman hash histogram

        Args:
            G (nx.Graph): graph to compte the histogram of

            node_label (str): node label to use as the starting node label of the
                Weisfeiler-Lehman hashing process

            n_iter (int): number of iterations of the Weisfeiler-Lehman algorithm
                to run

        Returns:
            Dict: dictionary of the format {hash_value: n_nodes}.
        """
        hash_iter_0 = dict(
            Counter(list(dict(G.nodes(self.node_label)).values()))
        )
        hashes = dict(
            Counter(
                flatten_lists(
                    list(
                        nx.weisfeiler_lehman_subgraph_hashes(
                            G,
                            node_attr=self.node_label,
                            iterations=self.n_iter,
                        ).values()
                    )
                )
            )
        )
        return hashes | hash_iter_0

    def diagonal(self, X, Y, K):
        if X == Y:
            return np.diag(K), np.diag(K)
        else:
            # Compute the diagonal of the self-similarity kernel matrix
            diag_X = [self.dot_product((elem, elem)) for elem in X]
            diag_Y = [self.dot_product((elem, elem)) for elem in Y]
            return diag_X, diag_Y

    def parallel_dot_product(
        self, lst: Iterable,
    ) -> Iterable:  # pragma: no cover
        """Computes the inner product of elements in lst.

        Args:
            lst (Iterable): Iterable to compute the inner product of.

        Returns:
            Iterable: computed inner products.
        """
        res = list()
        for x in lst:
            res.append(
                {
                    list(x.keys())[0]: [
                        list(x.values())[0][0],
                        self.dot_product(list(x.values())[0][1]),
                    ]
                }
            )
        return res

    def dot_product(self, dicts: Tuple) -> int:  # pragma: no cover
        """Computes the inner product of two dictionaries using common
        keys. This dramatically improves computation times when the number
        of keys is large but the overlap between the two dictionaries in
        the tuple is low.

        Args:
            dicts (Tuple): pair of dictionaries to compute the kernel from.

        Returns:
            int: dot product value of dicts
        """
        running_sum = 0
        # 0 * x = 0 so we only need to iterate over common keys
        for key in set(dicts[0].keys()).intersection(dicts[1].keys()):
            running_sum += dicts[0][key] * dicts[1][key]
        return running_sum

    def compute_gram_matrix(
        self, X: List[nx.Graph], Y: Union[List[nx.Graph], None] = None
    ) -> np.ndarray:
        """Computes the Gram matrix of the Weisfeiler-Lehman kernel.

        Args:
            X (List[nx.Graph]): List of graphs to use in the kernel.
            Y (Union[List[nx.Graph], None], optional): List of graphs to use in the kernel. Defaults to None.

        Returns:
            np.ndarray: _description_
        """

        def handle_hashes_single_threaded(
            X: Iterable[nx.Graph],
        ) -> Iterable[Dict]:  # pragma: no cover
            """Handles hashes when n_jobs = None on a single thread.

            Args:
                X (Iterable[nx.Graph]): Iterable of graphs to compute the hashes from

            Returns:
                Iterable[Dict]: hash histograms
            """
            X_hashed = list()
            for g in X:
                X_hashed.append(self.compute_wl_hashes(g))
            return X_hashed

        if not self.biased and self.normalize:
            self.biased = True
            if self.verbose:
                print("Setting biased=True to allow for normalization")

        check_wl_input(X)

        if Y is not None:
            check_wl_input(Y)

        if not self.precomputed:
            if Y == None:  # pragma: no cover
                Y = X

            if Y == X and self.n_jobs is not None:
                X_hashed = distribute_function(
                    self.compute_wl_hashes,
                    X,
                    self.n_jobs,
                    show_tqdm=self.verbose,
                    tqdm_label="Compute hashes of X",
                )
                Y_hashed = X_hashed
            elif X == Y and self.n_jobs is None:
                X_hashed = handle_hashes_single_threaded(X)
                Y_hashed = X_hashed
            elif X != Y and self.n_jobs is None:
                X_hashed = handle_hashes_single_threaded(X)
                Y_hashed = handle_hashes_single_threaded(Y)
            elif X != Y and self.n_jobs is not None:
                X_hashed = distribute_function(
                    self.compute_wl_hashes,
                    X,
                    n_jobs=self.n_jobs,
                    show_tqdm=self.verbose,
                    tqdm_label="Compute hashes of X",
                )
                Y_hashed = distribute_function(
                    self.compute_wl_hashes,
                    Y,
                    n_jobs=self.n_jobs,
                    show_tqdm=self.verbose,
                    tqdm_label="Compute hashes of Y",
                )
        else:
            X_hashed = X
            if X != Y and Y is not None:
                Y_hashed = Y
            else:
                Y_hashed = X_hashed

        # It's faster to process n_jobs lists than to have one list and
        # dispatch one item at a time.
        if not self.biased and X == Y:
            iters_data = list(list(combinations(X_hashed, 2)))
            iters_idx = list(combinations(range(len(X_hashed)), 2))
        elif self.biased and X == Y:
            iters_data = list(list(combinations_with_replacement(X_hashed, 2)))
            iters_idx = list(
                combinations_with_replacement(range(len(X_hashed)), 2)
            )
        else:
            iters_data = list(list(product(X_hashed, Y_hashed)))
            iters_idx = list(
                product(range(len(X_hashed)), range(len(Y_hashed)))
            )

        keys = generate_random_strings(10, len(flatten_lists(iters_data)))
        iters = [
            {key: [idx, data]}
            for key, idx, data in zip(keys, iters_idx, iters_data)
        ]
        if self.n_jobs is not None:
            iters = list(chunks(iters, self.n_jobs,))
            matrix_elems = flatten_lists(
                distribute_function(
                    self.parallel_dot_product,
                    iters,
                    self.n_jobs,
                    show_tqdm=self.verbose,
                    tqdm_label="Compute dot products",
                )
            )
        else:
            matrix_elems = self.parallel_dot_product(iters)

        K = np.zeros((len(X_hashed), len(Y_hashed)), dtype=int)
        for elem in matrix_elems:
            coords = list(elem.values())[0][0]
            val = list(elem.values())[0][1]
            K[coords[0], coords[1]] = val
        if X == Y:
            # mirror the matrix along diagonal
            K = np.triu(K) + np.triu(K, 1).T

        if self.normalize:
            # From https://github.com/ysig/GraKeL/blob/33ffff18d99c13f8afc0438a5691cb1206b119fb/grakel/kernels/weisfeiler_lehman.py#L300
            X_diag, Y_diag = self.diagonal(X_hashed, Y_hashed, K)
            K = np.nan_to_num(np.divide(K, np.sqrt(np.outer(X_diag, Y_diag))))
        return K



# -*- coding: utf-8 -*-
import contextlib
from collections import Counter
from itertools import product
from string import ascii_letters
from typing import Any, Callable, Dict, Iterable, List

import joblib
import networkx as nx
from joblib import Parallel, delayed
from tqdm import tqdm

from .exception import UniquenessError


def generate_random_strings(string_length: int, n_strings: int) -> List[str]:
    """Generates random strings

    Args:
        string_length (int): string length of the elements in the returned List.
        n_strings (int): number of strings to return

    Raises:
        UniquenessError: if the number of unique strings cannot be returned
            from the given string length.

    Returns:
        List[str]: unique strings
    """
    unique_strings = list()
    for idx, item in enumerate(product(ascii_letters, repeat=string_length)):
        if idx == n_strings:
            break
        unique_strings.append("".join(item))

    if len(unique_strings) != n_strings:
        raise UniquenessError(
            f"Cannot generate enough unique strings from given string length "
            "of {string_length}. Please increase the string_length to continue."
        )
    return unique_strings


def distribute_function(
    func: Callable,
    X: Iterable,
    n_jobs: int,
    tqdm_label: str = None,
    total: int = 1,
    show_tqdm: bool = True,
    **kwargs,
) -> Any:
    """Wraps a function to execute it across multiple threads.
    **kwargs are passed to the function to be parallelized.

    Args:
        func (Callable): function to be parallelized
        X (Iterable): data to use in func.
        n_jobs (int): number of threads to execute func
        tqdm_label (str, optional): Label of the tqdm progress bar. Defaults to None.
        total (int, optional): Total length of X if it cannot be deduced from X. Defaults to 1.
        show_tqdm (bool, optional): Whether or not to show tqdm. Defaults to True.

    Returns:
        Any: result of the parallel execution of func on X.
    """
    if total == 1:
        total = len(X)
    if show_tqdm:
        with tqdm_joblib(tqdm(desc=tqdm_label, total=total)) as progressbar:
            Xt = Parallel(n_jobs=n_jobs)(delayed(func)(x, **kwargs) for x in X)
    else:
        Xt = Parallel(n_jobs=n_jobs)(delayed(func)(x, **kwargs) for x in X)
    return Xt


def flatten_lists(lists: List) -> List:
    """Removes one level of nested list.
    i.e.

    flatten_lists([[1],[1,1,[1,1]]])
    becomes
    [1,1,1,[1,1]]

    Args:
        lists (list): list to remove one level of nestedness from.

    Returns:
        list: flattened list
    """
    result = list()
    for _list in lists:
        _list = list(_list)
        if _list != []:
            result += _list
        else:  # pragma: no cover
            continue
    return result


def chunks(lst: List, n: int) -> List:
    """Divides lst into n roughly equally-sized chunks.

    Args:
        lst (List): list to divide.
        n (int): number of chunks to divide lst

    Returns:
        List: lst divided into n roughly equally sized chunks.
    """
    k, m = divmod(len(lst), n)
    return (
        lst[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n)
    )


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar.

    Code stolen from https://stackoverflow.com/a/58936697
    """

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastwlk', 'fastwlk.utils']

package_data = \
{'': ['*']}

install_requires = \
['joblib>=1.1.0,<2.0.0',
 'networkx>=2.6.3,<3.0.0',
 'numpy>=1.22.1,<2.0.0',
 'tqdm>=4.62.3,<5.0.0']

setup_kwargs = {
    'name': 'fastwlk',
    'version': '0.2.13',
    'description': 'fastwlk is a Python package that implements a fast version of the Weisfeiler-Lehman kernel.',
    'long_description': "=============================\nFastWLK\n=============================\n\n.. image:: https://github.com/pjhartout/fastwlk/actions/workflows/main.yml/badge.svg\n        :target: https://github.com/pjhartout/fastwlk/\n\n\n.. image:: https://img.shields.io/pypi/v/fastwlk.svg\n        :target: https://pypi.python.org/pypi/fastwlk\n\n\n.. image:: https://codecov.io/gh/pjhartout/fastwlk/branch/main/graph/badge.svg?token=U054MJONED\n      :target: https://codecov.io/gh/pjhartout/fastwlk\n\n.. image:: https://img.shields.io/website-up-down-green-red/http/shields.io.svg\n   :target: https://pjhartout.github.io/fastwlk/\n\n\nQuick Links\n-------------------------\n`Documentation`_\n\n`Installation`_\n\n`Usage`_\n\n`Contributing`_\n\n\nWhat does ``fastwlk`` do?\n-------------------------\n\n\n``fastwlk`` is a Python package that implements a fast version of the\nWeisfeiler-Lehman kernel. It manages to outperform current state-of-the-art\nimplementations on sparse graphs by implementing a number of improvements\ncompared to vanilla implementations:\n\n1. It parallelizes the execution of Weisfeiler-Lehman hash computations since\n   each graph's hash can be computed independently prior to computing the\n   kernel.\n\n2. It parallelizes the computation of similarity of graphs in RKHS by computing\n   batches of the inner products independently.\n\n3. When comparing graphs, lots of computations are spent processing\n   positions/hashes that do not actually overlap between Weisfeiler-Lehman\n   histograms. As such, we manually loop over the overlapping keys,\n   outperforming numpy dot product-based implementations on collections of\n   sparse graphs.\n\nThis implementation works best when graphs have relatively few connections\ncompared to the number of possible connections and are reasonably dissimilar\nfrom one another. If you are not sure the graphs you are using are either sparse\nor dissimilar enough, try to benchmark this package with others out there using `this script`_.\n\nHow fast is ``fastwlk``?\n-------------------------\n\nRunning the benchmark script in ``examples/benchmark.py`` shows that for the\ngraphs in ``data/graphs.pkl``, we get an approximately 80% speed improvement\nover other implementations like `grakel`_. The example dataset contains 2-nn\ngraphs extracted from 100 random proteins from the human proteome from the\n`AlphaFold EBI database`_.\n\nTo see how much faster this implementation is for your use case:\n\n.. code-block:: console\n\n   $ git clone git://github.com/pjhartout/fastwlk\n   $ poetry install\n   $ poetry run python examples/benchmark.py\n\nYou will need to swap out the provided ``graphs.pkl`` with a pickled iterable of\ngraphs from the database you are interested in.\n\n.. _Documentation: https://pjhartout.github.io/fastwlk/\n.. _Installation: https://pjhartout.github.io/fastwlk/installation.html\n.. _Usage: https://pjhartout.github.io/fastwlk/usage.html\n.. _Contributing: https://pjhartout.github.io/fastwlk/contributing.html\n.. _grakel: https://github.com/ysig/GraKeL\n.. _AlphaFold EBI database: https://alphafold.ebi.ac.uk/download\n.. _this script: https://github.com/pjhartout/fastwlk/blob/main/examples/benchmark.py\n",
    'author': 'Philip Hartout',
    'author_email': 'philip.hartout@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
